"""
TCP Server to capture UDF script output.
Unlike HTTP Transport, we have to use real server with proper bindings.

Incoming connections may be blocked by firewalls!
It mostly affects local laptops trying to connect to Exasol located in remote data centre.

This module can work in two modes:

    1) DEBUG MODE
    Useful for manual debugging during UDF script development.
    Accepts connections from all VM's, but displays output of first connected VM only.
    Runs forever, until stopped by Ctrl + C (SIGTERM).

    How to run: python -m pyexasol_utils.script_output


    2) SCRIPT MODE
    Useful for production usage and during last stages of development.
    Accepts connections from all VM's and stores output into separate log files.
    Runs for one SQL statement only, stops automatically.

    How to run: ExaConnection.execute_with_udf_output()


We use ThreadingMixIn because:
a) Workload is pure I/O, so GIL should not be a problem.
b) Potential amount of VM's connected in parallel is huge and may surpass 1000+, which may cause problems with forks.

"""
import socket
import socketserver
import sys
import os
import shutil
import pathlib


class ExaScriptOutputServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    connected_clients = 0
    total_clients = 0

    # Stop all sub-threads immediately
    daemon_threads = True
    allow_reuse_address = True

    output_dir_path = None
    initial_ppid = None

    def get_output_address(self):
        return f"{socket.gethostbyname(socket.getfqdn())}:{self.socket.getsockname()[1]}"

    def service_actions(self):
        self.check_orphaned(self.initial_ppid)

    @staticmethod
    def check_orphaned(initial_ppid):
        """
        Raise exception if current process is "orphaned" (parent process is dead)
        It is useful to stop PyEXASOL socket servers from being stuck in process list after parent process was killed

        Currently it works only for POSIX OS
        Please let me know if you know a good way to detect orphans on Windows
        """
        current_ppid = os.getppid()

        if sys.platform != "win32" and initial_ppid and current_ppid != initial_ppid:
            raise RuntimeError(f"Current process is orphaned, initial ppid={initial_ppid}, current ppid={current_ppid}")


class ExaScriptOutputHandler(socketserver.StreamRequestHandler):
    def setup(self):
        super().setup()
        self.server.connected_clients += 1
        self.server.total_clients += 1

    def finish(self):
        super().finish()
        self.server.connected_clients -= 1


class ExaScriptOutputDebugModeHandler(ExaScriptOutputHandler):
    def handle(self):
        if self.server.connected_clients == 1:
            print('\n-------- NEW STATEMENT --------', flush=True)

            # Read and flush line-by-line, show log to user as soon as possible
            for line in self.rfile:
                sys.stdout.buffer.write(line)
                sys.stdout.buffer.flush()
        else:
            dst = open(os.devnull, 'wb')
            shutil.copyfileobj(self.rfile, dst)
            dst.close()


class ExaScriptOutputScriptModeHandler(ExaScriptOutputHandler):
    def handle(self):
        path = self.server.output_dir_path / (str(self.server.total_clients).rjust(5, '0') + '.log')
        dst = open(path, 'wb')

        shutil.copyfileobj(self.rfile, dst)
        dst.close()

    def finish(self):
        super().finish()

        # No more opened connections? -> Shutdown server
        if self.server.connected_clients == 0:
            self.server.shutdown()


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(prog=f'python -m pyexasol_utils.script_output')

    parser.add_argument('--host', default='0.0.0.0', help='Specific address to bind TCPServer (default: 0.0.0.0)')
    parser.add_argument('--port', default=0, help='Specific port to bind TCPServer (default: random port)', type=int)
    parser.add_argument('--output-dir', default=None, help='Directory to write enumerated log files, one file per VM')
    parser.add_argument('--ppid', default=0, help='PID of parent process', type=int)

    args = parser.parse_args()

    # Capture mode: collect output of all VM's into specific "output_dir"
    if args.output_dir:
        output_dir_path = pathlib.Path(args.output_dir)

        if not output_dir_path.is_dir():
            raise ValueError(f"Output_dir does not exist or not a directory: {output_dir_path}")

        # Start TCP server
        server = ExaScriptOutputServer((args.host, args.port), ExaScriptOutputScriptModeHandler)
        server.output_dir_path = output_dir_path
        server.initial_ppid = args.ppid

        # Send output address to the main process
        sys.stdout.buffer.write(f'{server.get_output_address()}\n'.encode())
        sys.stdout.buffer.flush()

        # Handle incoming connections
        # Server is stopped by shutdown() call in handler after closing last incoming connection
        server.serve_forever()
        server.server_close()

    # Debug mode: display output of one VM into terminal, discard output of other VM's
    else:
        # Start TCP server with debug handler
        server = ExaScriptOutputServer((args.host, args.port), ExaScriptOutputDebugModeHandler)

        # Send pre-generated SQL with output address to user terminal
        sys.stdout.write(f"ALTER SESSION SET SCRIPT_OUTPUT_ADDRESS = '{server.get_output_address()}';\n")
        sys.stdout.flush()

        # Stop server manually with SIGTERM (Ctrl + C, etc.) when debugging is finished
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            pass

        server.server_close()



"""
Extension with Exasol-specific helper functions
"""

from .exceptions import ExaRuntimeError


class ExaExtension(object):
    def __init__(self, connection):
        self.connection = connection
        self.reserved_words = None

    def get_columns(self, object_name):
        """
        DEPRECATED, please use `.meta.sql_columns` instead

        Get information about columns of table or view (Websocket format)
        Object name may be passed as tuple to specify custom schema
        """
        object_name = self.connection.format.default_format_ident(object_name)
        return self.get_columns_sql(f"SELECT * FROM {object_name}")

    def get_columns_sql(self, query, query_params=None):
        """
        DEPRECATED, please use `.meta.sql_columns` instead

        Get columns of SQL query without executing it (Websocket format)
        It relies on prepared statement which is closed immediately without execution
        """
        stmt = self.connection.cls_statement(self.connection, query, query_params, prepare=True)
        columns = stmt.columns()
        stmt.close()

        return columns

    def insert_multi(self, table_name, data, columns=None):
        """
        INSERT small number of rows into table using prepared statement
        It provides better performance for small data sets of 10,000 rows or less compared to .import_from_iterable()

        Please use .import_from_iterable() for larger data sets and better memory efficiency
        Please use .import_from_pandas() to import from data frame regardless of its size

        You may use "columns" argument to specify custom order of columns for insertion
        If some columns are not included in this list, NULL or DEFAULT value will be used instead
        """

        # Convert possible iterator into list
        data = list(data)

        if len(data) == 0:
            raise ExaRuntimeError(self.connection, "At least one row of data is required for insert_multi()")

        params = {
            'table_name': self.connection.format.default_format_ident(table_name),
            'columns': '',
            'values': ', '.join(['?'] * len(data[0]))
        }

        if columns:
            params['columns'] = f"({','.join([self.connection.format.default_format_ident(c) for c in columns])})"

        query = "INSERT INTO {table_name!r}{columns!r} VALUES ({values!r})"

        stmt = self.connection.cls_statement(self.connection, query, params, prepare=True)
        stmt.execute_prepared(data)
        stmt.close()

        return stmt

    def get_sys_columns(self, object_name):
        """
        DEPRECATED, please use `.meta.list_columns` instead

        Get information about columns of table or view (SYS format)
        Object name may be passed as tuple to specify custom schema
        """
        if isinstance(object_name, tuple):
            schema = self.connection.format.default_format_ident_value(object_name[0])
            object_name = self.connection.format.default_format_ident_value(object_name[1])
        else:
            schema = self.connection.current_schema()
            object_name = self.connection.format.default_format_ident_value(object_name)

        sql = """/*snapshot execution*/
            SELECT c.column_name, c.column_type, c.column_maxsize, c.column_num_scale,
                   c.column_is_nullable, c.column_is_distribution_key, c.column_default,
                   c.column_comment, t.type_name
            FROM EXA_ALL_COLUMNS c
                JOIN EXA_SQL_TYPES t ON (c.column_type_id=t.type_id)
            WHERE c.column_schema={schema}
                AND c.column_table={object_name}
            ORDER BY c.column_ordinal_position
        """

        st = self._execute(sql, {'schema': schema, 'object_name': object_name})
        res = list()

        for r in st:
            res.append({
                'name': r['column_name'],
                'type': r['type_name'],
                'sql_type': r['column_type'],
                'size': r['column_maxsize'],
                'scale': r['column_num_scale'],
                'nulls': r['column_is_nullable'],
                'distribution_key': r['column_is_distribution_key'],
                'default': r['column_default'],
                'comment': r['column_comment'],
            })

        return res

    def get_sys_tables(self, schema=None, table_name_prefix=''):
        """
        DEPRECATED, please use `.meta.list_tables` instead

        Get information about tables in selected schema(SYS format)
        Output may be optionally filtered by table name prefix
        """
        if schema is None:
            schema = self.connection.current_schema()
        else:
            schema = self.connection.format.default_format_ident_value(schema)

        table_name_prefix = self.connection.format.default_format_ident_value(table_name_prefix)
        table_name_prefix = self.connection.format.escape_like(table_name_prefix)

        sql = """/*snapshot execution*/
            SELECT *
            FROM EXA_ALL_TABLES
            WHERE table_schema={schema}
                AND table_name LIKE '{table_name_prefix!r}%'
            ORDER BY table_name ASC
        """

        st = self._execute(sql, {'schema': schema, 'table_name_prefix': table_name_prefix})
        res = list()

        for r in st:
            res.append({
                'table_name': r['table_name'].lower() if self.connection.options['lower_ident'] else r['table_name'],
                'table_schema': r['table_schema'].lower() if self.connection.options['lower_ident'] else r['table_schema'],
                'table_is_virtual': r['table_is_virtual'],
                'table_has_distribution_key': r['table_has_distribution_key'],
                'table_comment': r['table_comment'],
            })

        return res

    def get_sys_views(self, schema=None, view_name_prefix=''):
        """
        DEPRECATED, please use `.meta.list_views` instead

        Get information about views in selected schema(SYS format)
        Output may be optionally filtered by view name prefix
        """
        if schema is None:
            schema = self.connection.current_schema()
        else:
            schema = self.connection.format.default_format_ident_value(schema)

        view_name_prefix = self.connection.format.default_format_ident_value(view_name_prefix)
        view_name_prefix = self.connection.format.escape_like(view_name_prefix)

        sql = """/*snapshot execution*/
            SELECT *
            FROM EXA_ALL_VIEWS
            WHERE view_schema={schema}
                AND view_name LIKE '{view_name_prefix!r}%'
            ORDER BY view_name ASC
        """

        st = self._execute(sql, {'schema': schema, 'view_name_prefix': view_name_prefix})
        res = list()

        for r in st:
            res.append({
                'view_name': r['view_name'].lower() if self.connection.options['lower_ident'] else r['view_name'],
                'view_schema': r['view_schema'].lower() if self.connection.options['lower_ident'] else r['view_schema'],
                'scope_schema': r['scope_schema'].lower() if self.connection.options['lower_ident'] else r['scope_schema'],
                'view_text': r['view_text'],
                'view_comment': r['view_comment'],
            })

        return res

    def get_sys_schemas(self, schema_name_prefix=''):
        """
        DEPRECATED, please use `.meta.list_schemas` instead

        Get information about schemas (SYS format)
        Output may be optionally filtered by schema name prefix
        """
        schema_name_prefix = self.connection.format.default_format_ident_value(schema_name_prefix)
        schema_name_prefix = self.connection.format.escape_like(schema_name_prefix)

        sql = """/*snapshot execution*/
            SELECT *
            FROM EXA_SCHEMAS
            WHERE schema_name LIKE '{schema_name_prefix!r}%'
            ORDER BY schema_name ASC
        """

        st = self._execute(sql, {'schema_name_prefix': schema_name_prefix})
        res = list()

        for r in st:
            res.append({
                'schema_name': r['schema_name'].lower() if self.connection.options['lower_ident'] else r['schema_name'],
                'schema_owner': r['schema_owner'].lower() if self.connection.options['lower_ident'] else r['schema_owner'],
                'schema_is_virtual': r['schema_is_virtual'],
                'schema_comment': r['schema_comment'],
            })

        return res

    def get_reserved_words(self):
        """
        DEPRECATED, please use `.meta.list_sql_keywords` instead

        Get reserved keywords which cannot be used as identifiers without double-quotes
        Never hard-code this list! It changes with every Exasol versions
        """
        if self.reserved_words is None:
            sql = """
                SELECT keyword
                FROM EXA_SQL_KEYWORDS
                WHERE reserved IS TRUE
                ORDER BY keyword
            """

            self.reserved_words = self._execute(sql).fetchcol()

        return self.reserved_words

    def get_disk_space_usage(self):
        """
        Exasol still lacks standard function to measure actual disk space usage
        We're trying to mitigate this problem by making custom function
        """
        sql = """
            SELECT measure_time,
                   (committed_size * redundancy + temp_swap_data) AS occupied_size,
                   (device_size * redundancy + hdd_free) AS total_size
            FROM "$EXA_STATS_DB_SIZE"
            ORDER BY measure_time DESC
            LIMIT 1
        """

        row = self._execute(sql).fetchone()

        if row is None:
            return None

        row['occupied_size'] = int(row['occupied_size'])
        row['total_size'] = int(row['total_size'])

        row['free_size'] = row['total_size'] - row['occupied_size']
        row['occupied_size_percent'] = round(row['occupied_size'] / row['total_size'] * 100, 2)

        return row

    def export_to_pandas_with_dtype(self, query_or_table, query_params=None):
        """
        Export to pandas and attempt to guess correct dtypes based on Exasol columns
        Since pandas has significantly smaller range of allowed values, this function makes many assumptions
        Please use it as baseline for your own function for complex cases

        Small decimal -> int32
        Big decimal -> int64
        Double -> float64
        Date, Timestamp -> datetime64[ns]
        Everything else -> category (!)
        """

        if query_params:
            query_or_table = self.connection.format.format(query_or_table, **query_params)

        if isinstance(query_or_table, tuple) or str(query_or_table).strip().find(' ') == -1:
            columns = self.get_columns(query_or_table)
        else:
            columns = self.get_columns_sql(query_or_table)

        params = {
            'names': list(),
            'dtype': dict(),
            'parse_dates': list(),
            'na_values': dict(),
            'infer_datetime_format': True,
            'engine': 'c',
            'skip_blank_lines': False,
        }

        for k, c in columns.items():
            params['names'].append(k)

            if c['type'] == 'DATE':
                params['dtype'][k] = 'object'
                params['na_values'][k] = '0001-01-01'
                params['parse_dates'].append(k)

            elif c['type'] == 'TIMESTAMP':
                params['dtype'][k] = 'object'
                params['na_values'][k] = '0001-01-01 00:00:00'
                params['parse_dates'].append(k)

            elif c['type'] == 'DECIMAL':
                if c['scale'] > 0:
                    params['dtype'][k] = 'float64'
                else:
                    if c['precision'] <= 9:
                        params['dtype'][k] = 'int32'
                    else:
                        params['dtype'][k] = 'int64'

            elif c['type'] == 'DOUBLE':
                params['dtype'][k] = 'float64'

            else:
                params['dtype'][k] = 'category'

        def callback(pipe, dst, **kwargs):
            import pandas
            return pandas.read_csv(pipe, **kwargs)

        return self.connection.export_to_callback(callback, None, query_or_table, None, params)

    def explain_last(self, details=False):
        """
        Returns profiling information for last executed query
        This function should be called immediately after execute()

        details=False returns AVG or MAX values for all Exasol nodes
        details=True returns separate rows for each individual Exasol node (column "iproc")

        Details are useful to detect bad data distribution and imbalanced execution

        If you want to see real values of CPU, MEM, HDD, NET columns, please enable Exasol profiling first with:
        ALTER SESSION SET PROFILE = 'ON';
        """
        self._execute('FLUSH STATISTICS')

        sql = """
            SELECT part_id /* PyEXASOL explain_last */
                {iproc_col!r}
                , part_name
                , part_info
                , object_schema
                , object_name
                , object_rows
                , in_rows
                , out_rows
                , duration
                , start_time
                , stop_time
                , cpu
                , mem_peak
                , temp_db_ram_peak
                , hdd_read
                , hdd_write
                , net
                , remarks
            FROM {table_name!q}
            WHERE session_id=CURRENT_SESSION
                AND stmt_id = CURRENT_STATEMENT - {stmt_offset!d}
            ORDER BY {order_by!r}
        """

        params = {
            'iproc_col': ', iproc' if details else '',
            'table_name': '$EXA_PROFILE_DETAILS_LAST_DAY' if details else '$EXA_PROFILE_LAST_DAY',
            'order_by': 'part_id ASC, iproc ASC' if details else 'part_id ASC',
            'stmt_offset': 4 if self.connection.attr['autocommit'] else 2,
        }

        return self._execute(sql, params).fetchall()

    def _execute(self, query, query_params=None):
        # Preserve ext-functions output format regardless of current options for user queries
        options = {
            'fetch_dict': True,
            'fetch_mapper': None,
            'lower_ident': True,
        }

        return self.connection.cls_statement(self.connection, query, query_params, **options)

    def __repr__(self):
        return f'<{self.__class__.__name__} session_id={self.connection.session_id()}>'



import os
import re
import subprocess
import sys
import threading

HTTP_EXPORT = 'export'
HTTP_IMPORT = 'import'


class ExaSQLThread(threading.Thread):
    """
    Thread class which re-throws any Exception to parent thread
    """
    def __init__(self, connection, compression):
        self.connection = connection
        self.compression = compression

        self.params = {}
        self.http_proc = None
        self.exa_proxy_list = []
        self.exc = None

        super().__init__()

    def set_http_proc(self, http_proc):
        self.http_proc = http_proc

    def set_exa_proxy_list(self, exa_proxy_list):
        self.exa_proxy_list = exa_proxy_list if isinstance(exa_proxy_list, list) else [exa_proxy_list]

    def run(self):
        try:
            self.run_sql()
        except BaseException as e:
            self.exc = e

            # In case of SQL error terminate HTTP server
            # It closes other end of pipe and interrupts I/O in callback function in main thread
            if self.http_proc is not None:
                self.http_proc.terminate()

    def run_sql(self):
        pass

    def join_with_exc(self, *args):
        super().join(*args)

        if self.exc:
            raise self.exc

    def build_file_list(self):
        files = list()

        if 'format' in self.params:
            if self.params['format'] not in ['gz', 'bz2', 'zip']:
                raise ValueError(f"Unsupported compression format: {self.params['format']}")

            ext = self.params['format']
        else:
            ext = 'gz' if self.compression else 'csv'

        if self.connection.options['encryption']:
            prefix = 'https://'
        else:
            prefix = 'http://'

        csv_cols = self.build_csv_cols()

        for i, proxy in enumerate(self.exa_proxy_list):
            files.append(f"AT '{prefix}{proxy}' FILE '{str(i).rjust(3, '0')}.{ext}'{csv_cols}")

        return files

    def build_columns_list(self):
        if 'columns' not in self.params:
            return ''

        return f"({','.join([self.connection.format.default_format_ident(c) for c in self.params['columns']])})"

    def build_csv_cols(self):
        if 'csv_cols' not in self.params:
            return ''

        safe_csv_cols_regexp = re.compile(r"^(\d+|\d+\.\.\d+)(\sFORMAT='[^'\n]+')?$", re.IGNORECASE)

        for c in self.params['csv_cols']:
            if not safe_csv_cols_regexp.match(c):
                raise ValueError(f"Value [{c}] is not a safe csv_cols part")

        return f"({','.join(self.params['csv_cols'])})"


class ExaSQLExportThread(ExaSQLThread):
    """
    Build and run IMPORT query into separate thread
    Main thread is busy outputting data in callbacks
    """
    def __init__(self, connection, compression, query_or_table, export_params):
        super().__init__(connection, compression)

        self.query_or_table = query_or_table
        self.params = export_params

    def run_sql(self):
        if isinstance(self.query_or_table, tuple) or str(self.query_or_table).strip().find(' ') == -1:
            export_source = self.connection.format.default_format_ident(self.query_or_table)
        else:
            # New lines are mandatory to handle queries with single-line comments '--'
            export_query = self.query_or_table.lstrip(" \n").rstrip(" \n;")
            export_source = f'(\n{export_query}\n)'

            if self.params.get('columns'):
                raise ValueError("Export option 'columns' is not compatible with SQL query export source")

        parts = list()

        parts.append(f"EXPORT {export_source}{self.build_columns_list()} INTO CSV")
        parts.extend(self.build_file_list())

        if self.params.get('delimit'):
            delimit = str(self.params['delimit']).upper()

            if delimit != 'AUTO' and delimit != 'ALWAYS' and delimit != 'NONE':
                raise ValueError('Invalid value for export parameter DELIMIT: ' + delimit)

            parts.append(f"DELIMIT = {delimit}")

        if self.params.get('encoding'):
            parts.append(f"ENCODING = {self.connection.format.quote(self.params['encoding'])}")

        if self.params.get('null'):
            parts.append(f"NULL = {self.connection.format.quote(self.params['null'])}")

        if self.params.get('row_separator'):
            parts.append(f"ROW SEPARATOR = {self.connection.format.quote(self.params['row_separator'])}")

        if self.params.get('column_separator'):
            parts.append(f"COLUMN SEPARATOR = {self.connection.format.quote(self.params['column_separator'])}")

        if self.params.get('column_delimiter'):
            parts.append(f"COLUMN DELIMITER = {self.connection.format.quote(self.params['column_delimiter'])}")

        if self.params.get('with_column_names'):
            parts.append("WITH COLUMN NAMES")

        self.connection.execute("\n".join(parts))


class ExaSQLImportThread(ExaSQLThread):
    """
    Build and run EXPORT query into separate thread
    Main thread is busy parsing results in callbacks
    """
    def __init__(self, connection, compression, table, import_params):
        super().__init__(connection, compression)

        self.table = table
        self.params = import_params

    def run_sql(self):
        table_ident = self.connection.format.default_format_ident(self.table)

        parts = list()

        parts.append(f"IMPORT INTO {table_ident}{self.build_columns_list()} FROM CSV")
        parts.extend(self.build_file_list())

        if self.params.get('encoding'):
            parts.append(f"ENCODING = {self.connection.format.quote(self.params['encoding'])}")

        if self.params.get('null'):
            parts.append(f"NULL = {self.connection.format.quote(self.params['null'])}")

        if self.params.get('skip'):
            parts.append(f"SKIP = {self.connection.format.safe_decimal(self.params['skip'])}")

        if self.params.get('trim'):
            trim = str(self.params['trim']).upper()

            if trim != 'TRIM' and trim != 'LTRIM' and trim != 'RTRIM':
                raise ValueError('Invalid value for import parameter TRIM: ' + trim)

            parts.append(trim)

        if self.params.get('row_separator'):
            parts.append(f"ROW SEPARATOR = {self.connection.format.quote(self.params['row_separator'])}")

        if self.params.get('column_separator'):
            parts.append(f"COLUMN SEPARATOR = {self.connection.format.quote(self.params['column_separator'])}")

        if self.params.get('column_delimiter'):
            parts.append(f"COLUMN DELIMITER = {self.connection.format.quote(self.params['column_delimiter'])}")

        self.connection.execute("\n".join(parts))


class ExaHTTPProcess(object):
    """
    HTTP communication and compression / decompression is offloaded to sub-process
    It communicates with main process using pipes
    """
    def __init__(self, host, port, compression, encryption, mode):
        self.host = host
        self.port = port
        self.compression = compression
        self.encryption = encryption
        self.mode = mode

        self.server = None
        self.proxy = None
        self.proc = None

        self.read_pipe = None
        self.write_pipe = None

    def start(self):
        args = [sys.executable,
                '-m', 'pyexasol_utils.http_transport',
                '--host', self.host,
                '--port', str(self.port),
                '--mode', self.mode,
                '--ppid', str(os.getpid())
                ]

        if self.compression:
            args.append('--compression')

        if self.encryption:
            args.append('--encryption')

        self.proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, bufsize=0)
        self.proxy = self.proc.stdout.readline().decode().rstrip('\n')

        self.read_pipe = self.proc.stdout
        self.write_pipe = self.proc.stdin

    def join(self):
        if self.proc:
            self.read_pipe.close()
            self.write_pipe.close()

            return self.proc.wait()

        return None

    def join_with_exc(self):
        code = self.join()

        if code != 0:
            raise RuntimeError(f"HTTP transport process finished with exitcode: {code}")

        return code

    def terminate(self):
        if self.proc:
            self.proc.terminate()

    def get_proxy(self):
        if self.proxy is None:
            raise RuntimeError("Proxy 'host:port' string is not available")

        return self.proxy


class ExaHTTPTransportWrapper(object):
    """
    Start fake HTTP server, obtain proxy "host:port" string
    Send it to parent process

    Block into "export_to_callback()" or "import_from_callback()" call,
    wait for incoming connection, process data, exit.
    """
    def __init__(self, host, port, mode, compression=False, encryption=False):
        self.http_proc = ExaHTTPProcess(host, port, compression, encryption, mode)
        self.http_proc.start()

        self.proxy = self.http_proc.get_proxy()

    def get_proxy(self):
        return self.proxy

    def export_to_callback(self, callback, dst, callback_params=None):
        if not callable(callback):
            raise ValueError('Callback argument is not callable')

        if callback_params is None:
            callback_params = {}

        try:
            result = callback(self.http_proc.read_pipe, dst, **callback_params)

            self.http_proc.read_pipe.close()
            self.http_proc.join_with_exc()

            return result
        except Exception as e:
            # Close HTTP Server if it is still running
            self.http_proc.terminate()
            self.http_proc.join()

            raise e

    def import_from_callback(self, callback, src, callback_params=None):
        if not callable(callback):
            raise ValueError('Callback argument is not callable')

        if callback_params is None:
            callback_params = {}

        try:
            result = callback(self.http_proc.write_pipe, src, **callback_params)

            self.http_proc.write_pipe.close()
            self.http_proc.join_with_exc()

            return result
        except Exception as e:
            # Close HTTP Server if it is still running
            self.http_proc.terminate()
            self.http_proc.join()

            raise e

    def __repr__(self):
        return f'<{self.__class__.__name__} proxy={self.proxy}>'



from . import constant
from .exceptions import ExaRuntimeError


class ExaMetaData(object):
    """
    This class implements lock-free meta data requests using `/*snapshot execution*/` SQL hint described in IDEA-476
    https://www.exasol.com/support/browse/IDEA-476

    If you still get locks, please make sure to update Exasol server to the latest minor version

    This class also implements no SQL metadata commands introduced in Exasol v7.0 via .execute_meta_nosql() function
    """
    snapshot_execution_hint = '/*snapshot execution*/'

    def __init__(self, connection):
        self.connection = connection
        self.sql_keywords = None

    def sql_columns(self, query, query_params=None):
        """
        Get result set columns of SQL query without executing it
        """
        st = self.connection.cls_statement(self.connection, query, query_params, prepare=True)
        columns = st.columns()
        st.close()

        return columns

    def schema_exists(self, schema_name):
        object_name = self.connection.format.default_format_ident_value(schema_name)

        if self.connection.protocol_version() >= constant.PROTOCOL_V2:
            st = self.execute_meta_nosql('getSchemas', {
                'schema': object_name,
            })
        else:
            st = self.execute_snapshot("""
                SELECT 1
                FROM sys.exa_schemas
                WHERE schema_name={object_name}
            """, {
                'object_name': object_name,
            })

        return st.rowcount() > 0

    def table_exists(self, table_name):
        if isinstance(table_name, tuple):
            object_schema = self.connection.format.default_format_ident_value(table_name[0])
            object_name = self.connection.format.default_format_ident_value(table_name[1])
        else:
            object_schema = self.connection.current_schema()
            object_name = self.connection.format.default_format_ident_value(table_name)

        if self.connection.protocol_version() >= constant.PROTOCOL_V2:
            st = self.execute_meta_nosql('getTables', {
                'schema': object_schema,
                'table': object_name,
                'tableTypes': ['TABLE'],
            })
        else:
            st = self.execute_snapshot("""
                SELECT 1
                FROM sys.exa_all_tables
                WHERE table_schema={object_schema}
                    AND table_name={object_name}
            """, {
                'object_schema': object_schema,
                'object_name': object_name,
            })

        return st.rowcount() > 0

    def view_exists(self, view_name):
        if isinstance(view_name, tuple):
            object_schema = self.connection.format.default_format_ident_value(view_name[0])
            object_name = self.connection.format.default_format_ident_value(view_name[1])
        else:
            object_schema = self.connection.current_schema()
            object_name = self.connection.format.default_format_ident_value(view_name)

        if self.connection.protocol_version() >= constant.PROTOCOL_V2:
            st = self.execute_meta_nosql('getTables', {
                'schema': object_schema,
                'table': object_name,
                'tableTypes': ['VIEW'],
            })
        else:
            st = self.execute_snapshot("""
                SELECT 1
                FROM sys.exa_all_views
                WHERE view_schema={object_schema}
                    AND view_name={object_name}
            """, {
                'object_schema': object_schema,
                'object_name': object_name,
            })

        return st.rowcount() > 0

    def list_schemas(self, schema_name_pattern='%'):
        st = self.execute_snapshot("""
            SELECT *
            FROM sys.exa_schemas
            WHERE schema_name LIKE {schema_name_pattern}
            ORDER BY schema_name ASC
        """, {
            'schema_name_pattern': schema_name_pattern,
        })

        return st.fetchall()

    def list_tables(self, table_schema_pattern='%', table_name_pattern='%'):
        st = self.execute_snapshot("""
            SELECT *
            FROM sys.exa_all_tables
            WHERE table_schema LIKE {table_schema_pattern}
                AND table_name LIKE {table_name_pattern}
            ORDER BY table_schema ASC, table_name ASC
        """, {
            'table_schema_pattern': table_schema_pattern,
            'table_name_pattern': table_name_pattern,
        })

        return st.fetchall()

    def list_views(self, view_schema_pattern='%', view_name_pattern='%'):
        st = self.execute_snapshot("""
            SELECT *
            FROM sys.exa_all_views
            WHERE view_schema LIKE {view_schema_pattern}
                AND view_name LIKE {view_name_pattern}
            ORDER BY view_schema ASC, view_name ASC
        """, {
            'view_schema_pattern': view_schema_pattern,
            'view_name_pattern': view_name_pattern,
        })

        return st.fetchall()

    def list_columns(self, column_schema_pattern='%', column_table_pattern='%'
                     , column_object_type_pattern='%', column_name_pattern='%'):
        st = self.execute_snapshot("""
            SELECT *
            FROM sys.exa_all_columns
            WHERE column_schema LIKE {column_schema_pattern}
                AND column_table LIKE {column_table_pattern}
                AND column_object_type LIKE {column_object_type_pattern}
                AND column_name LIKE {column_name_pattern}
        """, {
            'column_schema_pattern': column_schema_pattern,
            'column_table_pattern': column_table_pattern,
            'column_object_type_pattern': column_object_type_pattern,
            'column_name_pattern': column_name_pattern,
        })

        return st.fetchall()

    def list_objects(self, object_name_pattern='%', object_type_pattern='%', owner_pattern='%', root_name_pattern='%'):
        st = self.execute_snapshot("""
            SELECT *
            FROM sys.exa_all_objects
            WHERE object_name LIKE {object_name_pattern}
                AND object_type LIKE {object_type_pattern}
                AND owner LIKE {owner_pattern}
                AND root_name LIKE {root_name_pattern}
        """, {
            'object_name_pattern': object_name_pattern,
            'object_type_pattern': object_type_pattern,
            'owner_pattern': owner_pattern,
            'root_name_pattern': root_name_pattern,
        })

        return st.fetchall()

    def list_object_sizes(self, object_name_pattern='%', object_type_pattern='%', owner_pattern='%', root_name_pattern='%'):
        st = self.execute_snapshot("""
            SELECT *
            FROM sys.exa_all_object_sizes
            WHERE object_name LIKE {object_name_pattern}
                AND object_type LIKE {object_type_pattern}
                AND owner LIKE {owner_pattern}
                AND root_name LIKE {root_name_pattern}
        """, {
            'object_name_pattern': object_name_pattern,
            'object_type_pattern': object_type_pattern,
            'owner_pattern': owner_pattern,
            'root_name_pattern': root_name_pattern,
        })

        return st.fetchall()

    def list_indices(self, index_schema_pattern='%', index_table_pattern='%', index_owner_pattern='%'):
        st = self.execute_snapshot("""
            SELECT *
            FROM sys.exa_all_indices
            WHERE index_schema LIKE {index_schema_pattern}
                AND index_table LIKE {index_table_pattern}
                AND index_owner LIKE {index_owner_pattern}
        """, {
            'index_schema_pattern': index_schema_pattern,
            'index_table_pattern': index_table_pattern,
            'index_owner_pattern': index_owner_pattern,
        })

        return st.fetchall()

    def list_sql_keywords(self):
        """
        Get reserved SQL keywords which cannot be used as identifiers without double-quote escaping
        Never hardcode this list! It might change with next Exasol server version without warning
        """
        if not self.sql_keywords:
            if self.connection.protocol_version() >= constant.PROTOCOL_V2:
                st = self.execute_meta_nosql('getKeywords')

                self.sql_keywords = [r['KEYWORD'] for r in st.fetchall() if r['RESERVED'] is True]
            else:
                st = self.execute_snapshot("""
                    SELECT keyword
                    FROM EXA_SQL_KEYWORDS
                    WHERE reserved IS TRUE
                    ORDER BY keyword
                """)

                self.sql_keywords = st.fetchcol()

        return self.sql_keywords

    def execute_snapshot(self, query, query_params=None):
        """
        Execute query in snapshot transaction mode using SQL hint
        fetch_dict=True is enforced to prevent users from relying on order of columns in system views
        """
        options = {
            'fetch_dict': True,
        }

        return self.connection.cls_statement(self.connection, f"{self.snapshot_execution_hint}{query}", query_params, **options)

    def execute_meta_nosql(self, meta_command, meta_params=None):
        """
        Execute no SQL meta data command introduced in Exasol 7.0+
        This feature requires WebSocket protocol v2 or higher

        List of available commands: https://github.com/exasol/websocket-api/blob/master/docs/WebsocketAPIV2.md#metadata-related-commands
        """
        if self.connection.protocol_version() < constant.PROTOCOL_V2:
            raise ExaRuntimeError(self.connection, 'Protocol version 2 is required to execute nosql meta data commands')

        # Security check, prevents execution of dangerous commands if meta_command argument is dynamic
        if meta_command[0:3] != 'get':
            raise ExaRuntimeError(self.connection, 'Meta command name should start with prefix \'get*\'')

        options = {
            'fetch_dict': True,
        }

        return self.connection.cls_statement(self.connection, meta_command, meta_params, meta_nosql=True, **options)

    def __repr__(self):
        return f'<{self.__class__.__name__} session_id={self.connection.session_id()}>'
